/* cts_dump - a userspace tool to dump information present in conntrack
 * state sync packets */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include <errno.h>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>

#include <linux/config.h>
#include <linux/netfilter_ipv4/ct_sync.h>

/* cut n' pasted from kernel includes... */
#define NIPQUAD(addr) \
        ((unsigned char *)&addr)[0], \
        ((unsigned char *)&addr)[1], \
        ((unsigned char *)&addr)[2], \
        ((unsigned char *)&addr)[3]

#define DUMP_TUPLE(tp)					\
printf("tuple %u %u.%u.%u.%u:%hu -> %u.%u.%u.%u:%hu\n",	\
       (tp)->dst.protonum,				\
       NIPQUAD((tp)->src.ip), ntohs((tp)->src.u.all),	\
       NIPQUAD((tp)->dst.ip), ntohs((tp)->dst.u.all))

static int prepare_socket(char *interface)
{
	int fd;
	struct ifreq ifr;
	struct ip_mreqn mreq;
	struct sockaddr_in sa;

	fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);

	memset(&sa, 0, sizeof(sa));
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = INADDR_ANY;
	sa.sin_port = htons(1999);

	bind(fd, (struct sockaddr *)&sa, sizeof(sa));

	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, interface, IFNAMSIZ);
	ioctl(fd, SIOCGIFINDEX, &ifr);
	
	memset(&mreq, 0, sizeof(mreq));
	mreq.imr_ifindex = ifr.ifr_ifindex;
	inet_aton("224.0.0.82", &mreq.imr_multiaddr);
	printf("binding to %s, ifindex %d\n", inet_ntoa(mreq.imr_multiaddr),
		mreq.imr_ifindex);
	fflush(stdout);
	setsockopt(fd, SOL_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq));

	return fd;
}

static void dump_update_conntrack(struct ct_sync_msghdr *msgh)
{
	struct ct_sync_conntrack *sct;

	sct = (struct ct_sync_conntrack *)(msgh + 1);

	printf("     orig  ");
	DUMP_TUPLE(&sct->orig);
	printf("     reply ");
	DUMP_TUPLE(&sct->reply);
	printf("     CT status %04x expires %ld helper '%s' expecting %d\n",
	       sct->status, sct->expires, sct->helper, sct->expecting);
	printf("     NAT initialized %x num_manips %d helper '%s'\n",
	       sct->nat_initialized, sct->nat_num_manips, sct->nat_helper);
}

static void dump_delete_conntrack(struct ct_sync_msghdr *msgh)
{
	struct ip_conntrack_tuple *t;

	t = (struct ip_conntrack_tuple *)(msgh + 1);

	printf("     ");
	DUMP_TUPLE(t);
}

static void handle_msg(struct ct_sync_msghdr *msgh, int num)
{
	char *res, *type;

	switch (msgh->resource) {
		case CT_SYNC_RES_CONNTRACK:
			res = "CONNTRACK";
			break;
		case CT_SYNC_RES_EXPECT:
			res = "EXPECT";
			break;
		default:
			res = "HARALD_IS_STUPID";
			break;
	}

	switch (msgh->type) {
		case CT_SYNC_MSG_UPDATE:
			type = "UPDATE";
			break;
		case CT_SYNC_MSG_DELETE:
			type = "DELETE";
			break;
		default:
			type = "HARALD_IS_STUPID";
			break;
	}
			
	printf("  %2u R:%s T:%s L:%u\n", num, res, type, ntohs(msgh->len));

	switch (msgh->type) {
	case CT_SYNC_MSG_UPDATE:
		switch (msgh->resource) {
		case CT_SYNC_RES_CONNTRACK:
			dump_update_conntrack(msgh);
			break;
		default:
			break;
		}
		break;
	case CT_SYNC_MSG_DELETE:
		switch (msgh->resource) {
		case CT_SYNC_RES_CONNTRACK:
			dump_delete_conntrack(msgh);
			break;
		default:
			break;
		}
		break;
	default:
		break;
	}
}

static int handle_packet(const char *recvbuf, int rlen, const struct sockaddr_in *sa)
{
	struct ct_sync_pkthdr *pkth = (struct ct_sync_pkthdr *) recvbuf;
	struct ct_sync_msghdr *msgh;
	char *type;

	printf("%s:%u -> ", inet_ntoa(sa->sin_addr), ntohs(sa->sin_port));

	if (rlen < sizeof(*pkth)) {
		printf("short packet\n");
		return -1;
	}

	switch (pkth->pkttype) {

	case CT_SYNC_PKT_MASTER_ANNOUNCE:
		type = "MASTER_ANNOUNCE";
		break;
	case CT_SYNC_PKT_SLAVE_INITSYNC:
		type = "SLAVE_INITSYNC";
		break;
	case CT_SYNC_PKT_SYNC:
		type = "SYNCDATA";
		break;
	case CT_SYNC_PKT_NACK:
		type = "NACK";
		break;
	}

	printf("V:%u T:%s S:%u N:%u msgs=%u len=%u\n",
		pkth->version, type, ntohs(pkth->pktseq), 
		pkth->node_id, pkth->count, rlen);

	if (pkth->pkttype == CT_SYNC_PKT_SYNC) {
		int i = 0;
		int msg_ofs = sizeof(*pkth);
		for (msgh = (void *)pkth + sizeof(*pkth); 
		     msg_ofs < rlen;
		     msgh = (void *)pkth + msg_ofs) {
			handle_msg(msgh, i);
			msg_ofs += sizeof(*msgh) + ntohs(msgh->len);
			i++;
		}
	}

	fflush(stdout);
	return 0;
}

int main(int argc, char **argv)
{
	int fd;
	char recvbuf[4000];
	struct sockaddr_in sa;
	socklen_t sl = sizeof(sa);

	if (argc != 2) {
		fprintf(stderr, "USAGE: cts_dump interface\n");
		return 1;
	}

	fd = prepare_socket(argv[1]);

	while (1) {
		int len;
		len = recvfrom(fd, recvbuf, 4000, 0, (struct sockaddr *)&sa, 
			 	&sl);
		if (len < 0) {
			fprintf(stderr, "ERROR %s", strerror(errno));
		} 
		handle_packet(recvbuf, len, &sa);
	}

	return 0;
}

